#include "stdafx.h"

const char * m_pchDefaultRegistryKey = "DataConfigDefault";

bool g_bDefaultLoaded=false;		
CDataConfigure g_oDefaultConfig;

const COLORREF g_aTraceColor[MAX_CHANNELS] = { 
		RGB(  0,  0,  0), 
		RGB(  0,  0, 64), 
		RGB(  0,  0,128), 
		RGB(  0,  0,192), 
		 RGB(  0,  0,255), 
		 RGB(  0, 64,  0), 
		 RGB(  0, 64, 64), 
		 RGB(  0, 64,128), 
		RGB(  0, 64,192), 
		RGB(  0, 64,255), 
		RGB(  0,128,  0), 
		RGB(  0,128, 64), 
		 RGB(  0,128,128), 
		 RGB(  0,128,192), 
		 RGB(  0,128,255), 
		 RGB(  0,192,  0)
		};

//=============================================================================
CEventList :: CEventList(  ) { 
	m_iEvents=0;
	m_iEventsAlloc=0;
	m_aiClass=NULL;
	m_aiDuration=NULL;
	m_aiStimIndex=NULL;
	m_iStimChannels=0;
	m_aiStimChans=NULL;
	m_afStimulii=NULL;
	m_afBuffer=NULL;
	m_iStimAlloc=0;
	m_iStimCount=0;
	}
	
CEventList :: ~CEventList() {
	if( m_aiClass!=NULL ) 
		delete [] m_aiClass;
	if( m_aiDuration!=NULL )
		delete [] m_aiDuration;
	if( m_aiStimIndex!=NULL )
		delete [] m_aiStimIndex;
	if( m_aiStimChans!=NULL )
		delete [] m_aiStimChans;
	if( m_afStimulii!=NULL )
		delete [] m_afStimulii;
	if( m_afBuffer!=NULL ) 
		delete [] m_afBuffer;
	}
		
bool CEventList :: Allocate( int iEvents, bool bStim ) { 
	m_aiClass = new short[iEvents];
	m_aiDuration = new int[iEvents];
	if( m_aiClass==NULL || m_aiDuration==NULL  ) 
		return false;
	m_iEventsAlloc = iEvents;
	if( bStim ) { 
		m_aiStimIndex = new unsigned short[iEvents];
		if( m_aiStimIndex==NULL ) 
			return false;
		}
	return true;
	}

bool CEventList :: AllocateStimChan( int iStimChans ) { 
	if( iStimChans>0 ) { 
		m_afBuffer = new float[iStimChans];
		m_afStimulii = new float[EVTLST_ALLOCSIZE];
		m_aiStimChans = new int[iStimChans];
		if( m_afStimulii==NULL || m_aiStimChans==NULL )
			return false;
		m_iStimAlloc = EVTLST_ALLOCSIZE;
		m_iStimChannels = 0;
		}
	return true;
	}

bool CEventList :: AddStimChan( int iCh ) { 
	if( m_iStimChannels >= m_iStimAlloc ) 
		return false;
	m_aiStimChans[m_iStimChannels++]=iCh;
	return true;
	}

bool CEventList :: AddEvent( int iDuration, short iClass ) { 
	if( m_iEvents==m_iEventsAlloc ) 
		return false;
	m_aiClass[m_iEvents]=iClass;
	m_aiDuration[m_iEvents]=iDuration;
	if( m_aiStimIndex != NULL ) 
		m_aiStimIndex[m_iEvents]=0;
	m_iEvents++;
	return true;
	}

bool CEventList :: CombineEvent( int iDuration, short iClass ) { 
	if( m_iEvents>0 ) 
		m_aiDuration[m_iEvents-1]+=iDuration;
	return true;
	}

int CEventList :: FindOrAddStim( ) { 
	//-- if no stimulii
	if( m_afStimulii==NULL ) 
		return -1;

	//-- if already in list 
    for( int i=0; i<m_iStimCount; ++i ) 
		if( memcmp( m_afBuffer, m_afStimulii+m_iStimChannels*i, m_iStimChannels*sizeof(float) ) == 0 ) 
			return i;

	//-- if list is full 
	if( (m_iStimCount+1)*m_iStimChannels >= m_iStimAlloc ) { 
		//-- add another chunk
		float * pfNew = new float[m_iStimAlloc+EVTLST_ALLOCSIZE];
		if( pfNew == NULL ) 
			return -1;
		memcpy( pfNew, m_afStimulii, sizeof(float)*m_iStimCount*m_iStimChannels );
		delete [] m_afStimulii;
		m_afStimulii = pfNew;
		m_iStimAlloc+=EVTLST_ALLOCSIZE;
		}

	//-- add to list at end.
	memcpy( m_afStimulii+m_iStimChannels*m_iStimCount, m_afBuffer, m_iStimChannels*sizeof(float) );
	return m_iStimCount++;
	}

//=============================================================================
int VisibleChannels( CItemArray * paChannels ) { 
	int iVisChannels=0, iCh, iChannels=paChannels->Count();
	for( iCh=0; iCh<iChannels; ++iCh ) 
		if( ((CChannel *)paChannels->GetAt(iCh))->m_bVisible ) 
			++iVisChannels;
	return iVisChannels;
	}

//=============================================================================
CChannel :: CChannel( ) { 
	m_cUnits="V";
	m_cName="Channel";
	m_bVisible=true;
	m_bBaseline=false;
	m_bFilter=false;
	m_dFilterKhz=0;
	m_bAutoScale=true;
	m_cr = g_aTraceColor[0];
	m_dChScaling = 1.0;
	m_dLowScale = -50.0f;
	m_dHighScale = 50.0f;
	m_dMin = 0.0;
	m_dMax = -1.0;
	}

const CChannel & CChannel :: operator=( const CChannel & oSource ) { 
	m_cUnits=oSource.m_cUnits;
	m_cName=oSource.m_cName;
	m_bVisible=oSource.m_bVisible;
	m_bBaseline=oSource.m_bBaseline;
	m_bFilter=oSource.m_bFilter;
	m_dFilterKhz=oSource.m_dFilterKhz;
	m_bAutoScale=oSource.m_bAutoScale;
	m_cr=oSource.m_cr;
	m_dChScaling=oSource.m_dChScaling;
	// these will not really neeed to be copied - they get recalculated.
	m_dLowScale=oSource.m_dLowScale;
	m_dHighScale=oSource.m_dHighScale;		
	m_dMin=oSource.m_dMin;
	m_dMax=oSource.m_dMax;
	return * this;
	}

CString CChannel :: ToString() {
	char buf[64];		// (only 22 used) "1YYYY999.999rrrgggbbb"
	sprintf( buf, "1%c%c%c%c%7.3f%3d%3d%3d", 
			m_bAutoScale?'Y':'N', m_bVisible?'Y':'N', m_bBaseline?'Y':'N', 
			m_bFilter?'Y':'N', m_dFilterKhz, 
			GetRValue(m_cr), GetGValue(m_cr), GetBValue(m_cr) );
	return CString( buf );
	}	

void CChannel :: FromString( const char * pchBuf ) { 
	int iLen = strlen(pchBuf);
	if( iLen>=21 && pchBuf[0]=='1' ) { 
		CBuffer o( strlen(pchBuf) );
		o.Fill( iLen, pchBuf );
		m_bAutoScale=(pchBuf[1]=='Y');
		m_bVisible=(pchBuf[2]=='Y');
		m_bBaseline=(pchBuf[3]=='Y');
		m_bFilter=(pchBuf[4]=='Y');
		m_dFilterKhz = atof_l( pchBuf+5, 7 );
		m_cr = RGB( atoi_l(pchBuf+12,3), atoi_l(pchBuf+15,3), atoi_l(pchBuf+18,3) );
		}
	}

double CChannel :: VisibleMax() { 
	return m_bAutoScale ? m_dMax : m_dHighScale ; 
	}
	
double CChannel :: VisibleMin(){ 
	return m_bAutoScale ? m_dMin : m_dLowScale ; 
	}

//=============================================================================
void CChanSegDisp :: DeleteDisp( bool bInit ) { 
	m_iDispHeight=0;
	m_iDispLength=0;
	if( ! bInit && m_piDispLow != NULL )
		delete [] m_piDispLow;
	m_piDispLow=NULL;

	if( ! bInit && m_piDispHigh != NULL )
		delete [] m_piDispHigh;
	m_piDispHigh=NULL;
	}

bool CChanSegDisp :: AllocateDisp( int iLen, bool bClear ) { 
	DeleteDisp();
	if( (m_piDispLow=new short[iLen])==NULL ) 
		return false;
	if( (m_piDispHigh=new short[iLen])==NULL ) {
		delete[]m_piDispLow;
		return false;
		}
	if( bClear ) { 
		memset( m_piDispLow, 0, iLen * sizeof(short) );
		memset( m_piDispHigh, 0, iLen * sizeof(short) );
		}
	m_iDispLength=iLen;
	return true;				
	}
	
CChanSegDisp :: CChanSegDisp( ) { 
	DeleteDisp(true);
	}
	
CChanSegDisp :: ~CChanSegDisp() { 
	DeleteDisp();
	}

//-- Draw the data to the display bitmap
bool CChanSegDisp :: Render( HDC dc, int iBtm, int iOffset, int iWid, COLORREF cr, bool bBaseline ) { 
	//TODO - draw baselines separately, then draw all baselines before any data 

	//-- Draw 1 data trace to bitmap
	HPEN penTrace = CreatePen( 0, 0, cr );
	HPEN penWas  = (HPEN) SelectObject( dc, penTrace );

	if( bBaseline )				// TODO - consider a dotted pen and draw with LineTo()
		for( int i=0; i<iWid && i+iOffset<m_iDispLength; i+=3 )		// +3 : dotted line
			SetPixel( dc, i, iBtm-m_iDispBase, cr );		

	for( int i=0; i<iWid && i+iOffset<m_iDispLength; ++i ) { 
		if( m_piDispLow[iOffset+i] == m_piDispHigh[iOffset+i] ) 
			SetPixel( dc, i, iBtm-m_piDispLow[iOffset+i], cr );
		else { 
			MoveToEx( dc, i, iBtm-m_piDispHigh[iOffset+i],NULL );
			LineTo( dc, i, iBtm-m_piDispLow[iOffset+i] );
			}
		}
				
	SelectObject( dc, penWas );
	DeleteObject( penTrace );
	
	return true;	
	}

//--- add IDL information to the display bitmap 	
bool CChanSegDisp :: RenderDispIdl( HDC dc, CIdlList * pIdl, int iBtm, int iOffset, double dPixelsPerSample ) { 
	int iDwell, iDwells=pIdl->m_aIdlDwells.Count();
	int iClass, iClasses=pIdl->m_aIdlClasses.Count();
	
	//-- create array of pixel amplitudes for classes 
	short * aiAmpPixel = new short[iClasses];
	if( aiAmpPixel == NULL ) 
		return false;
	for( iClass=0; iClass<iClasses; ++iClass )
		aiAmpPixel[iClass]=YPixel( pIdl->IdlClass(iClass)->m_dAmp );
		
	//-- 
	HPEN penTrace = CreatePen( 0, 0, pIdl->m_crColor );
	HPEN penWas = (HPEN) SelectObject( dc, penTrace );

	CIdlDwell * pDwell;
	int iDwellOffset = pIdl->m_iOffset;
	int iX,iY;
	
	for( iDwell=0; iDwell<iDwells; ++iDwell ) { 
		pDwell = pIdl->IdlDwell(iDwell);
		//-- start of dwell		
		iY = iBtm-aiAmpPixel[ pDwell->m_iClass ];
		iX = int(dPixelsPerSample * iDwellOffset) -iOffset;
		if( iDwell==0 ) 
			MoveToEx( dc, iX, iY, NULL );
		else 
			LineTo( dc, iX, iY );			// vertical bar connecting to next dwell
		//-- end of dwell
		iDwellOffset += pDwell->m_iDuration;
		iX = int(dPixelsPerSample * iDwellOffset) - iOffset;
		LineTo( dc, iX, iY );
		}
		
	SelectObject( dc, penWas );
	DeleteObject( penTrace );
	delete [] aiAmpPixel;

	return true;	
	} 
	
	
short CChanSegDisp :: YPixel( double dVal ) { 
	return (short) ((max( m_dMin, min( m_dMax, dVal)) - m_dMin )/m_dRange*m_iDispHeight  );	// height on pixel map of this pt.
	}

bool CChanSegDisp :: CreateDisp( double * pdData, int iDispLength, int iSamples, int iSkip, double dPixelsPerSample, int iHeight, CChannel * pCh ) { 
	//-- already created for correct dimensions, exit
	if( m_iDispHeight == iHeight && m_iDispLength == iDispLength ) 
		return true;
	
	//-- create arrays for display 
	AllocateDisp( iDispLength );
	m_iDispHeight = iHeight;
	
	//-- determine signal range displayed
	m_dMin = pCh->VisibleMin(); 
	m_dMax = pCh->VisibleMax();
	m_dRange = m_dMax-m_dMin;
	if( m_dRange==0.0 )
		return true;
	
	//-- Scan data and fill in disp
	short iVal;
 	int iPixel, iSetPixel=-1;
	for( int i=0; i<iSamples; ++i ) { 
		iVal = YPixel( pdData[i*iSkip] );
		iPixel = int(dPixelsPerSample * (i+1));		// pixel this displays at 
		if( iSetPixel<iPixel ) {				// 1st set for this pixel
			while( iSetPixel<iPixel ) {			// while - since there may be >1 pixel per sample
				iSetPixel++;
				m_piDispLow[iSetPixel] = iVal;
				m_piDispHigh[iSetPixel] = iVal;
				}
			}
		else {									// combine with previous
			m_piDispLow[iSetPixel] = min( iVal, m_piDispLow[iSetPixel] );
			m_piDispHigh[iSetPixel] = max( iVal, m_piDispHigh[iSetPixel] );
			}
		}
	
	//-- Display level for baseline	
	m_iDispBase = short((0.0-m_dMin) / m_dRange * iHeight);

	//-- repair gaps
	for( int i=1; i<iDispLength; ++i ) { 
		if( m_piDispLow[i-1] > m_piDispHigh[i] )
			m_piDispHigh[i] = m_piDispLow[i-1] = ( m_piDispLow[i-1] + m_piDispHigh[i] ) / 2;
		else if( m_piDispHigh[i-1] < m_piDispLow[i] ) 
			m_piDispLow[i] = m_piDispHigh[i-1] = ( m_piDispHigh[i-1] + m_piDispLow[i] ) / 2;
		}
		
	return true;	
	}

double CChanSegDisp :: AbsMax() { 
	return max( abs(m_dMin), m_dMax);
	}

//=============================================================================
CIdlList :: CIdlList() { 
	m_crColor=RGB(255,0,0); 
	m_iChannel=0;
	m_iOffset=0;
	m_iLength=0;
	}

CIdlClass * CIdlList :: IdlClass( int iClass ){
	return (CIdlClass *) (m_aIdlClasses[iClass]);
	}
	
CIdlDwell * CIdlList :: IdlDwell( int iDwell ){
	return (CIdlDwell *) (m_aIdlDwells[iDwell]);
	}

void CIdlList :: AllocateClasses( int iClasses ) { 
	while( m_aIdlClasses.Count()<iClasses ) 
		m_aIdlClasses.AddItem( new CIdlClass );
	}
	
void CIdlList :: AllocateDwells( int iDwells ) { 
	while( m_aIdlDwells.Count()<iDwells ) 
		m_aIdlDwells.AddItem( new CIdlDwell );
	}

void CIdlList :: AddDwell( int iClass, int iDuration, int iState ) { 
	CIdlDwell * pDwell = new CIdlDwell;
	pDwell->m_iClass=iClass;
	pDwell->m_iDuration=iDuration;
	pDwell->m_iState=iState;
	m_aIdlDwells.AddItem( pDwell );
	} 

//=============================================================================
CDigoutEvent :: CDigoutEvent( int iLine, int iStart ) { 
	m_iLine = iLine;
	m_iSampleStart = iStart;
	m_iSampleStop = 0;
	}
	
CDigoutEvent * CDigoutEventList :: Event( int iEvent ) { 
	if( iEvent<0 || iEvent>=m_aEvents.Count() ) 
		return NULL;
	return (CDigoutEvent *) m_aEvents[iEvent];
	}
	
CDigoutEventList :: CDigoutEventList() {
	for( int i=0; i<32; ++i ) 
		m_aiOn[i]=-1;
	m_iActiveLines=0;
	}
	
void CDigoutEventList :: AddEvent( int iLine, int iStart ) { 
	CDigoutEvent * pEvent = new CDigoutEvent( iLine, iStart );
	m_aiOn[iLine]=m_aEvents.Count();
	m_aEvents.AddItem( pEvent );
	}

void CDigoutEventList :: StopEvent( int iLine, int iStop ) { 
	CDigoutEvent * pEvent = Event( m_aiOn[iLine] );
	if( pEvent != NULL )
		pEvent->m_iSampleStop=iStop;
	if( iLine>=0 && iLine<32 ) 
		m_aiOn[iLine]=-1;
	}

int CDigoutEventList :: EventCount() {
	return m_aEvents.Count();
	}
	
void CDigoutEventList :: SetMasks( int iActive, int iAutoReset ) { 
	m_iAutoResetMask=iAutoReset;
	
	if( m_iActiveMask!=iActive ) {
		m_iActiveMask=iActive;
		//-- count set bits in m_iActiveMask - work up group in 2's, 4's, 8's, 16.
		//-- avoid repetitive calculation  : if input is same as prior call, return prior output.
		m_iActiveLines=0;
		unsigned int iBits = m_iActiveMask;
		iBits = ( iBits & 0x55555555) + (( iBits & 0xAAAAAAAA ) >>  1);
		iBits = ( iBits & 0x33333333) + (( iBits & 0xCCCCCCCC ) >>  2);
		iBits = ( iBits & 0x0f0f0f0f) + (( iBits & 0xf0f0f0f0 ) >>  4);
		iBits = ( iBits & 0x00ff00ff) + (( iBits & 0xff00ff00 ) >>  8);
		iBits = ( iBits & 0x0000ffff) + (( iBits & 0xffff0000 ) >> 16);
		m_iActiveLines = iBits;
		}
	else 
		m_iActiveLines = 0;
	
	m_iActiveLines = min(16,m_iActiveLines);		//TODO - remove this 		
	}
	
void CDigoutEventList :: LoadArray( int iLen, unsigned int * aiData ) { 
	m_aEvents.DeleteAll();
	SetMasks( aiData[0], aiData[1] );
	for( int i=2; i<iLen-1; i+=2 ) { 
		int iSample = aiData[i];
		int iLine = aiData[i+1] & 0xFFFF;	// but really 0x0003f would suffice for 0..31
		int iType = aiData[i+1] >> 16;
		if( iType==0 )	
			StopEvent( iLine, iSample );
		else if( iType==1 ) 
			AddEvent( iLine, iSample );
		else { 
			AddEvent( iLine, iSample );
			StopEvent( iLine, iSample+1 );
			}
		}
	}

//=============================================================================
CSegment :: CSegment() {
	m_dtStartTime=0.0;
	m_iPos=0;
	m_pdData=NULL;
	m_piDigoutDisp=NULL;
	m_bDigout=false;
	m_iSamples=0;
	m_iChannels=0;
	m_bInterleaved=false;
	m_iChanged=0;
	}

CSegment :: ~CSegment() { 
	if( m_piDigoutDisp != NULL )	
		delete [] m_piDigoutDisp;
	if( m_pdData != NULL )	
		delete [] m_pdData;
	}

CChanSegDisp * CSegment :: ChanSegDisp( int iCh ) { 
	return (CChanSegDisp *)(m_aChanSegDisps[iCh]);
	}

bool CSegment :: AllocateData( bool bClear ) { 
	DeleteData();
	if( (m_pdData=new double[m_iSamples*m_iChannels]) == NULL ) {
		m_iChannels = m_iSamples = 0;
		return false;
		}
	if( bClear ) 
		memset( m_pdData, 0, m_iSamples*m_iChannels*sizeof(double) );
	return true;
	}
	
void CSegment :: DeleteData( ) { 
	if( m_pdData != NULL )
		delete [] m_pdData;
	m_pdData=NULL;
	}

bool CSegment :: SetData( double * pData, int iSamples, int iChannels, bool bInterleaved ) { 
	m_iSamples=iSamples;
	m_iChannels=iChannels;
	m_bInterleaved=bInterleaved;
	DeleteData();
	m_pdData=pData;
	return true;
	}

bool CSegment :: CopyData( double * pData, int iSamples, int iChannels, bool bInterleaved ) { 
	m_iSamples=iSamples;
	m_iChannels=iChannels;
	m_bInterleaved=bInterleaved;
	if( ! AllocateData( ) ) 
		return false;
	memcpy( m_pdData, pData, iSamples*iChannels*sizeof(double) );
	return true;
	}

bool CSegment :: LoadDataDouble( int iSamples, int iChannels, FILE * h, 
							bool bInterleaved, double * adMin, double * adMax  ) { 
	m_bInterleaved=bInterleaved;
	if( ! AllocateData( ) ) 
		return false;
	if( fseek( h, m_iPos, SEEK_SET ) != 0 
			|| fread( m_pdData, sizeof(double), m_iSamples, h ) != m_iSamples ) { 
		DeleteData();
		return false;
		}
	if( adMin != NULL && iSamples>0 ) 
		for( int iChannel=0; iChannel<iChannels; ++iChannel ) { 
			double dVal = m_pdData[ bInterleaved ? iChannel : iChannel*iSamples ];
			adMax[iChannel]=adMin[iChannel]=dVal;
			for( int iSample=1; iSample<iSamples; ++iSample ) { 
				dVal = m_pdData[ bInterleaved ? iSample*iChannels+iChannel : iChannel*iSamples+iSample ];
				adMin[iChannel]=min(dVal,adMin[iChannel]);
				adMax[iChannel]=max(dVal,adMax[iChannel]);
				}
			}
				
	return true;
	}

bool CSegment :: LoadDataShort( int iSamples, int iChannels, FILE * h, double * adScaling, 
					double dFileScaling, bool bInterleaved, double * adMin, double * adMax ) { 
	m_bInterleaved=true;
	if( ! AllocateData( ) ) 
		return false;

	//----- allocate buffer of shorts to read into 
	short * piData = new short[iChannels*iSamples];
	if( piData==NULL ) 
		return false;

	//----- read data from file 
	int iRead = fread( piData, iChannels*sizeof(short), iSamples, h );
	if(  iRead != iSamples ) { 
		delete [] piData;
		return false;
		}
		
	//----- convert ints to doubles and reorganize the data ( time consuming )  
	double dVal, dScaling;
	if( adMin!=NULL || adMax!=NULL ) { 
		for( int iChannel=0; iChannel<iChannels; ++iChannel ) {
			dScaling = adScaling[iChannel]/dFileScaling; 
			adMin[iChannel] = adMax[iChannel] = 
				dScaling * piData[ bInterleaved ? iChannel : iChannel*iSamples ];
			for( int iSample=0; iSample<iSamples; ++iSample ) {
				dVal = dScaling * piData[ bInterleaved ? iSample*iChannels+iChannel : iChannel*iSamples+iSample ];
				m_pdData[iSample*iChannels+iChannel] = dVal;
				adMin[iChannel]=min(dVal,adMin[iChannel]);
				adMax[iChannel]=max(dVal,adMax[iChannel]);
				}
			}
		}
	else {
		for( int iChannel=0; iChannel<iChannels; ++iChannel ) { 
			dScaling = adScaling[iChannel]/dFileScaling; 
			for( int iSample=0; iSample<iSamples; ++iSample )
				m_pdData[iSample*iChannels+iChannel] = dScaling * piData[ bInterleaved ? iSample*iChannels+iChannel : iChannel*iSamples+iSample ];
			}
		}

	//----- Done 
	delete [] piData;
	return true;
	}

bool CSegment :: CreateDisp( int iTraceHeight, double dPixPerSamp, bool bOverlap, bool bDigout, CItemArray * paChannels) { 
	//-- count visible channels if not overlap to determine height per channel
	int iVisChannel=0, iVisChannels=bOverlap?1:max(1,VisibleChannels(paChannels));

	//----- CreateDisp for each channel 
	CChanSegDisp * pCh;
	int iChannelHeight, iChannels=paChannels->Count();
	int iDigoutHeight = 0;
	if( bDigout ) {
		if( iTraceHeight > 6*m_oDigoutList.m_iActiveLines ) 
			iDigoutHeight = 2;
		else if( iTraceHeight > 3*m_oDigoutList.m_iActiveLines ) 
			iDigoutHeight = 1;
		}
		
	//-- calc total length
	int iDispLength = int( dPixPerSamp * m_iSamples) + 1;
	
	for( int iCh=0; iCh<iChannels && iCh<m_iChannels; ++iCh ) {
		//-- add ChSeg to list if not already there
		if( iCh >= m_aChanSegDisps.Count() ) 
			m_aChanSegDisps.AddItem( new CChanSegDisp() );
			
		if( ((CChannel*)paChannels->GetAt(iCh))->m_bVisible ) { 
			++iVisChannel;

			//-- calc height of each channel ( all full height if overlap mode ) 	
			pCh = ChanSegDisp(iCh);
			if( bOverlap  )
				iChannelHeight = iTraceHeight-iDigoutHeight*m_oDigoutList.m_iActiveLines; 
			else	//-- allocation of all bits to channels (i.e. 10 bits / 3 channels = { 3, 4, 3 }
				iChannelHeight = MulDiv( iVisChannel, iTraceHeight-iDigoutHeight*m_oDigoutList.m_iActiveLines, iVisChannels ) 
								- MulDiv( iVisChannel-1, iTraceHeight-iDigoutHeight*m_oDigoutList.m_iActiveLines, iVisChannels );

			if( ! pCh->CreateDisp(
								m_pdData+(m_bInterleaved?iCh:m_iSamples*iCh), // address of 1st sample 
								iDispLength,
								m_iSamples, 
								m_bInterleaved ? m_iChannels : 1,	// iSkip between samples
								dPixPerSamp,	
								iChannelHeight, 
								(CChannel*)paChannels->GetAt(iCh) ) )
			return false;
			}
		}


	//------------------- Fill in digital display data
	if( m_bDigout ) { 
 		m_iDigoutHeight = iDigoutHeight;
 		m_iDigoutDispLength = 0;

		//-- allocate empty space
		if( m_piDigoutDisp != NULL )
			delete [] m_piDigoutDisp;
		if( (m_piDigoutDisp=new unsigned int[iDispLength])==NULL ) 
			return false;
		m_iDigoutDispLength = iDispLength;
		memset( m_piDigoutDisp, 0, iDispLength * sizeof(unsigned int) );

		//-- create mapping from line # to display index
		int aiMap[32], iLine=0;
		for( int i=0; i<32; ++i )				
			aiMap[i] = (m_oDigoutList.m_iActiveMask & (1<<i)) ? iLine++ : 31 ;

		//-- for each event, set bit in sample range 
		int iEvent, iEvents=m_oDigoutList.EventCount();
		int iPixel, iPixel0, iPixel1, iMask;
		for( iEvent=0; iEvent<iEvents; ++iEvent ) {
			CDigoutEvent * pEvent = m_oDigoutList.Event(iEvent);
			iPixel0 = int( dPixPerSamp * pEvent->m_iSampleStart );
			iPixel1 = int( min(iDispLength-1, dPixPerSamp*pEvent->m_iSampleStop) );
			iMask = ( 1<< aiMap[pEvent->m_iLine] );
			for( iPixel=iPixel0; iPixel<=iPixel1; ++iPixel ) 
				m_piDigoutDisp[iPixel] |= iMask;
			}
		}

	return true;
	}

void CSegment :: SetRanges( CItemArray * paChannels ) { 
	double dVal;
	
	if( m_pdData == NULL || m_iSamples==0 ) 
		return;
		
	int iChannels = paChannels->Count();
	for( int iCh=0; iCh<iChannels; ++iCh ) { 
		CChannel * pCh = (CChannel *)paChannels->GetAt(iCh);
		if( pCh->m_dMin > pCh->m_dMax )
			pCh->m_dMin = pCh->m_dMax = m_pdData[ m_bInterleaved ? iCh : m_iSamples*iCh ];
			
		for( int iSample=0; iSample<m_iSamples; ++iSample ) { 
			dVal = m_pdData[ m_bInterleaved ? iChannels*iSample+iCh : iCh*m_iSamples+iSample ];
			pCh->m_dMin = min( dVal, pCh->m_dMin );
			pCh->m_dMax = max( dVal, pCh->m_dMax );
			}
		}
	}

bool CSegment :: Render( HDC dc, bool bOverlap, bool bDispIdl, double dPixPerSamp, int iTop, int iOffset, int iWid, CItemArray * paChannels, COLORREF * aDigoutColors ) { 
	//-- Draw all visible data traces to bitmap 
	CChanSegDisp * pChanSegDisp;
	CChannel * pChannel;
	int iBtm=iTop;
	int iVisChannels = bOverlap ? 1 : max(1,VisibleChannels(paChannels));
	int iIdl, iIdls=m_aIdl.Count();
	
	//----- Render each segment-channel display 
	for( int iCh=0; iCh<m_aChanSegDisps.Count() && iCh<m_iChannels; ++iCh ) {
		pChannel = (CChannel*)paChannels->GetAt(iCh);
		if( pChannel->m_bVisible ) {
			pChanSegDisp = ChanSegDisp(iCh);
			if( bOverlap ) 
				iBtm = iTop+pChanSegDisp->m_iDispHeight;
			else 
				iBtm += pChanSegDisp->m_iDispHeight;
		
			if( ! pChanSegDisp->Render( dc, iBtm, iOffset, iWid, pChannel->m_cr, pChannel->m_bBaseline ) ) 
				return false;
				
			if( bDispIdl ) 				
				for( iIdl=0; iIdl<iIdls; ++iIdl )
					if( IdlList(iIdl)->m_iChannel==iCh ) 
						pChanSegDisp->RenderDispIdl( dc, (CIdlList *)(m_aIdl[iIdl]), iBtm, iOffset, dPixPerSamp );
			}
		}
		
	//----- Render Digital outputs 
	if( m_bDigout ) { 
		int iX, iBit, iY, iLines = m_oDigoutList.m_iActiveLines;
		
		//-- fill bgr for digout display
		RECT rct;
		SetRect( &rct, 0, iBtm, m_iDigoutDispLength, iBtm+m_iDigoutHeight*(iLines+1) );
		HBRUSH br = CreateSolidBrush( RGB(160,160,128) );
		FillRect( dc, &rct, br );
		DeleteObject( br );
		
		//-- draw digout traces
		COLORREF cr;
		for( iX=0; iX<iWid && iOffset+iX<m_iDigoutDispLength; ++iX )
			for( iBit=0; iBit<iLines; ++iBit ) {
				cr = aDigoutColors[iBit];
				if( m_piDigoutDisp[iOffset+iX] & (1<<iBit) ) 
					for( iY=0; iY<m_iDigoutHeight; ++iY ) 
						SetPixel( dc, iX, iBtm+m_iDigoutHeight*iBit+iY, cr );
				}
				
		}	
	return true;
	}
	
int CSegment :: TraceCount( int iSamplesPerTrace )  { 
	return 1+int((max(1,m_iSamples)-1)/iSamplesPerTrace);
	}

CIdlList * CSegment :: IdlList(int iIdlList) { 
	return (CIdlList *)(m_aIdl[iIdlList]);
	}

void CSegment :: DeleteOverlappingIdls( int iStart, int iStop ) { 
	if( iStop==-1 ) 
		iStop=m_iSamples-1;
	int iIdls = m_aIdl.Count();
	for( int iIdl=iIdls-1; iIdl>=0; --iIdl ) { 
		CIdlList * pIL = IdlList(iIdl);
		if( pIL->m_iOffset+pIL->m_iLength>=iStart && pIL->m_iOffset<=iStop )
			m_aIdl.DeleteItem(iIdl);
		}	
	}
		
//== 
bool CSegment :: CalculateSum( int iChStart, int iChStop, int iPtStart, int iPtStop, double * adSum, bool bInitToZero ) {
	int iPt, iPts=iPtStop-iPtStart+1;
	int iCh, iChs=iChStop-iChStart+1;
	//-- Zero initial values
	if( bInitToZero ) 
		for( iCh=0; iCh<iChs; ++iCh )
			adSum[iCh]=0.0;
		
	//-- Sum all pts
	double * pD;
	if( m_bInterleaved ) 
		for( iPt=iPtStart; iPt<=iPtStop; ++iPt ){ 
			pD=&(m_pdData[iPt*iChs]);
			for( iCh=iChStart; iCh<=iChStop; ++iCh ) 
				adSum[iCh-iChStart] += pD[iCh];
			}
	else 
		for( iCh=iChStart; iCh<=iChStop; ++iCh ) {
			pD=&(m_pdData[iCh*iPts]);
			for( iPt=iPtStart; iPt<=iPtStop; ++iPt ) 
				adSum[iCh-iChStart] += pD[iPt];
			}
	return true;
	}

bool CSegment :: CalculateVariance( int iChStart, int iChStop, int iPtStart, int iPtStop, double * adAvg, double * adVar, bool bInitToZero ) {
	int iPt, iPts=iPtStop-iPtStart+1;
	int iCh, iChs=iChStop-iChStart+1;
	//-- Zero initial values
	if( bInitToZero ) 
		for( iCh=0; iCh<iChs; ++iCh )
			adVar[iCh]=0.0;

	//-- Sum Variances
	double * pD;
	if( m_bInterleaved ) 
		for( iPt=iPtStart; iPt<=iPtStop; ++iPt ){ 
			pD=&(m_pdData[iPt*iChs]);
			for( iCh=iChStart; iCh<=iChStop; ++iCh ) 
				adVar[iCh] += (pD[iCh]-adAvg[iCh])*(pD[iCh]-adAvg[iCh]);
			}
	else 
		for( iCh=iChStart; iCh<=iChStop; ++iCh ) {
			pD=&(m_pdData[iCh*iPts]);
			for( iPt=iPtStart; iPt<=iPtStop; ++iPt ) 
				adVar[iCh] += (pD[iPt]-adAvg[iCh])*(pD[iPt]-adAvg[iCh]);
			}
	return true;
	}

int CSegment :: IDLEventList( int iChannel, int iChannels, 
			int iRanges, int * aiRangeStart, int * aiRangeStop, double dSampling, 
			double dDeadTime, CEventList * pEventList ) {
	int iEvents=0;
	int iPrevClass=-1;
	int iDur;

// TODO - for initial sizing calls, combining end of last seg with start of next will not be noticed.
	if( pEventList!=NULL && pEventList->m_iEvents>0 ) 
		iPrevClass = pEventList->m_aiClass[pEventList->m_iEvents-1];
	
	//-- for each Idl list 
	int iIdls = m_aIdl.Count();
	for( int iIdl=0; iIdl<iIdls; ++iIdl ) { 
		//-- if it overlaps the desired area 
		CIdlList * pIdl = IdlList( iIdl );
		if( pIdl->m_iChannel==iChannel 
				&& pIdl->m_iOffset<=aiRangeStop[iRanges-1]
				&& pIdl->m_iOffset+pIdl->m_iLength>=aiRangeStart[0] ) { 

			//-- for each dwell 
			int iPt=pIdl->m_iOffset;
			int iDwells = pIdl->m_aIdlDwells.Count();
			for( int iDwell=0; iDwell<iDwells && iPt<=aiRangeStop[iRanges-1]; ++iDwell ) {
				CIdlDwell * pDwell = pIdl->IdlDwell(iDwell);
				//-- check if this dwell overlaps any range 
				iDur=0;
				bool bFound=false;
				for( int iRange=0; iRange<iRanges; ++iRange ) 
					if( iPt+pDwell->m_iDuration > aiRangeStart[iRange] && iPt<=aiRangeStop[iRange] ) { 
						bFound=true;
						iDur = max( iDur, min( iPt+pDwell->m_iDuration, aiRangeStop[iRange] ) - max( iPt, aiRangeStart[iRange] ) );
						}
				
				//-- count this dwell (or portion of it)
				if( bFound ) { 
					//-- if shorter than dead time or same class as last event : 
					if( iPrevClass==pDwell->m_iClass || dSampling*iDur<dDeadTime ) { 
						//-- combine with previous event
						if( pEventList!=NULL ) 
							pEventList->CombineEvent( iDur, iPrevClass );
						}
					else { 
						iEvents++;
						if( pEventList != NULL ) 
							pEventList->AddEvent( iDur, pDwell->m_iClass );
						iPrevClass=pDwell->m_iClass;
						}
					}
				iPt += pDwell->m_iDuration;
				}
			}
		}

	if( pEventList != NULL && pEventList->m_iStimChannels>0 ) { 
		//-- for all new events
		int iDur, iOffset=aiRangeStart[0];
		float fSum;
		int iListEvents = pEventList->m_iEvents;
		for( int iEvent=iListEvents-iEvents; iEvent<iListEvents; ++iEvent ) { 
			iDur = pEventList->m_aiDuration[iEvent];
			iOffset += iDur;
			//-- calculate stimulii averages across event
			for( int iCh=0; iCh<pEventList->m_iStimChannels; ++iCh ) { 
				int iChannel = pEventList->m_aiStimChans[iCh];
				fSum=0.0f;
				for( int iSamp=iOffset-iDur; iSamp<iOffset; ++iSamp ) 
					fSum += float(m_pdData[iSamp*iChannels+iChannel]);
				pEventList->m_afBuffer[iCh] = fSum/iDur;
				}

			//-- add to list and set index for event 
			pEventList->m_aiStimIndex[iEvent] = pEventList->FindOrAddStim( );
			}
		}

	return iEvents;
	}

//=============================================================================
CDataConfigure :: CDataConfigure() { 	
	m_iTracesPerPage=6;				// 'Configure' parameters - edit in dialogue, persist to disk.
	m_iPointsPerTrace=10000;
	m_bOverlap=false;
	m_bDispLists=true;	
	m_bDispListLabels=true;	
	m_bDispCursors=true;	
	m_bDispFitCurves=true;
	m_bDispIdealizations=true;
	m_bDispDigout=true;	
	for( int i=0; i<DIGOUTLINES; ++i )
		m_acrDigout[i]=RGB((i%4)*85,255-(i%7)*42,0);//RGB(85*((i+3)%4),40+100*(i%3),32*(i%7));
	}

//----- add  1 channel, set it to defaults if available
bool CDataConfigure :: AddChannel( bool bDefault ) { 
	int iCh = m_aChannels.Count();
	if( iCh>=MAX_CHANNELS ) 
		return false;
		
	CChannel * pChannel = new CChannel(  );
	if( pChannel==NULL ) 
		return false;
	pChannel->m_cr=g_aTraceColor[iCh];
	
	if( bDefault && iCh<g_oDefaultConfig.ChannelCount() ) 
		*pChannel=*g_oDefaultConfig.Channel(iCh);
		
	m_aChannels.AddItem( pChannel );
	return true;
	}


void CDataConfigure :: CopyDefaultDataConfig( ) {
	m_iTracesPerPage=g_oDefaultConfig.m_iTracesPerPage;
	m_iPointsPerTrace=g_oDefaultConfig.m_iPointsPerTrace;
	m_bOverlap=g_oDefaultConfig.m_bOverlap;
	m_bDispLists=g_oDefaultConfig.m_bDispLists;
	m_bDispListLabels=g_oDefaultConfig.m_bDispListLabels;
	m_bDispCursors=g_oDefaultConfig.m_bDispCursors;
	m_bDispFitCurves=g_oDefaultConfig.m_bDispFitCurves;
	m_bDispIdealizations=g_oDefaultConfig.m_bDispIdealizations;
	m_bDispDigout=g_oDefaultConfig.m_bDispDigout;
	for( int iLine=0; iLine<DIGOUTLINES; ++iLine ) 
		m_acrDigout[iLine]=g_oDefaultConfig.m_acrDigout[iLine];
	//-- but do not copy channels - they will create from default when added
	}
	
void CDataConfigure :: LoadDefaultDataConfig() { 
	CString cLoad;
	if( ! g_bDefaultLoaded ) {
		cLoad = RegistryLoadString( m_pchDefaultRegistryKey );
		if( cLoad.Length()==0 ) 
			return;
		g_bDefaultLoaded=g_oDefaultConfig.FromStringDataConfig(&cLoad);
		}
	}
	
void CDataConfigure :: SetDefaultDataConfig() { 
	CString c = ToString();
	g_oDefaultConfig.FromStringDataConfig( &c );
	RegistrySaveString( m_pchDefaultRegistryKey, c.ConstBuffer() );
	}

CString CDataConfigure :: ToString(  ) { 
	CBuffer o;
	o.AppendInt(1);						// version # 1
	o.AppendInt( m_iTracesPerPage ); 
	o.AppendInt( m_iPointsPerTrace ); 
	o.AppendInt( (m_bOverlap ? 1 : 0 )
				+ (m_bDispLists ? 2 : 0)
				+ (m_bDispListLabels ? 4 : 0)
				+ (m_bDispCursors ? 8 : 0)
				+ (m_bDispFitCurves ? 16 : 0)
				+ (m_bDispIdealizations ? 32 : 0)
				+ (m_bDispDigout ? 64 : 0) );
	//-- per digital channel - color 
	o.AppendInt( DIGOUTLINES );
	for( int i=0; i<DIGOUTLINES; ++i ) 
		o.AppendInt( (int)(m_acrDigout[i]) );
		
	//-- per analog channel - all except name, units
	o.AppendInt( ChannelCount() );
	for( int i=0; i<ChannelCount(); ++i )
		o.AppendString( Channel(i)->ToString().ConstBuffer() );
		
	return o.ToString();
	}
	
bool CDataConfigure :: FromStringDataConfig( CString * pc ) { 
	CBuffer o;
	o.Fill( pc->Length(), pc->ConstBuffer() );
	
	int iVer, iFlags, iLines, iCR;
	if( ! o.ExtractInt(&iVer)
		|| ! o.ExtractInt(&m_iTracesPerPage)
		|| ! o.ExtractInt(&m_iPointsPerTrace)
		|| ! o.ExtractInt(&iFlags)
		|| ! o.ExtractInt(&iLines) ) 
		return false;
	m_bOverlap = (iFlags&1)>0;
	m_bDispLists  = (iFlags&2)>0;
	m_bDispListLabels  = (iFlags&4)>0;
	m_bDispCursors  = (iFlags&8)>0;
	m_bDispFitCurves = (iFlags&16)>0;
	m_bDispIdealizations = (iFlags&32)>0;
	m_bDispDigout = (iFlags&64)>0;

	//-- per digital channel - color 
	for( int i=0; i<DIGOUTLINES && i<iLines; ++i ) {
		o.ExtractInt(&iCR);
		m_acrDigout[i] = (COLORREF)iCR;
		}

	//-- per analog channel - all except name, units
	int iSavedChannels;
	CString cChannel;
	o.ExtractInt( &iSavedChannels );
	m_aChannels.DeleteAll();
	for( int i=0; i<iSavedChannels; ++i )
		if( o.ExtractString( &cChannel ) && i<MAX_CHANNELS )	{		// if able to get data 
			while( i>=ChannelCount()  )
				AddChannel(false);
			Channel(i)->FromString( cChannel.ConstBuffer() );
			}
			 
	return true;
	}

CChannel * CDataConfigure :: Channel( int iChannel ) { 
	return (CChannel *)(m_aChannels[iChannel]);
	}
	
int CDataConfigure :: ChannelCount() { 
	return m_aChannels.Count();	
	}

//=============================================================================
//-- constructor 
CData :: CData() { 
	m_oConfig.LoadDefaultDataConfig();
	m_oConfig.CopyDefaultDataConfig();
	m_pDataFile=NULL;
	ClearUpdated();
	m_iStatus=0;
	}
	
//-- destructor 
CData :: ~CData() { 
	if( m_pDataFile!=NULL )
		m_pDataFile->Close();
	}

void CData :: ClearUpdated() { 
	m_bDataUpdated=false;
	m_bListsUpdated=false;
	m_bCursorsUpdated=false;
	m_bFitCurvesUpdated=false;
	m_bIdlUpdated=false;
	m_bConfigUpdated=false;
	}

//-- Accessor to segment array 
CSegment * CData :: Segment( int iSeg ) { 
	return (CSegment *)m_aSegments[iSeg];
	}

//-- Accessor to List array 
CSelList * CData :: List( int iSeg ) { 
	return (CSelList *)m_aLists[iSeg];
	}

CSelList * CData :: List( const char * pchListName ) { 
	for( int iList=m_aLists.Count()-1; iList>=0; --iList ) 
		if( List(iList)->m_cName==pchListName )
			return List(iList);
	return NULL;
	}

//-- Open a .qdf file 
bool CData :: OpenQDF( const char * pchFileName ) { 
	//----- Open disk file, Load Data 
	m_pDataFile = new CQDSFInterface;
	if( ! m_pDataFile->Open( pchFileName, &m_dSampling, &m_dScaling, &m_oConfig, &m_aSegments, &m_aLists ) ) 
		return false;

	CQDSFFile * pQDSF = &(((CQDSFInterface*)m_pDataFile)->m_oQDSF);
		
	ClearUpdated();
	return true;			
	}			

bool CData :: WriteQDF( const char * pchFileName ) {
	if( m_pDataFile==NULL ) {							//----- Create a new .QDF file 
		// If dataFile object is null, create a new file and write everything to it.
		m_pDataFile = new CQDSFInterface;
		m_pDataFile->CreateData( pchFileName, m_dSampling, m_dScaling, &m_oConfig, &m_aSegments, &m_aLists );
		}
	else if( strcmp(m_pDataFile->m_cFileName.ConstBuffer(),pchFileName)!=0 ){	//----- If datafile name is changed, delete the interface object so a new file is created.
		// TODO - make sure everything is loaded to memory first.
		m_pDataFile->Close();
		m_pDataFile->CreateData( pchFileName, m_dSampling, m_dScaling, &m_oConfig, &m_aSegments, &m_aLists );
		}
	else {												//----- Update the existing datafile 
		m_pDataFile->SaveData( pchFileName, m_dSampling, m_dScaling, &m_oConfig, &m_aSegments, &m_aLists );
		}

	ClearUpdated();
	return true;
	}
	
bool CData :: SetStatus( int iStatus ) { 
	// TODO - MUTEX this
	if( iStatus != 0 && m_iStatus != 0 ) 
		return false;
	m_iStatus=iStatus;
	return true;
	}

int CData :: GetStatus() { 
	return m_iStatus;
	}
	
bool CData :: CalculateStatistics( int iChStart, int iChStop, int iSegStart, int iPtStart, int iSegStop, int iPtStop, int * piPoints, double * adAvg, double * adStd) {
	int iCh, iSeg, iPt0, iPt1, iPoints=0, iChs = iChStop-iChStart+1;

	for( iSeg=iSegStart; iSeg<=iSegStop; ++iSeg ) { 
		CSegment * pSeg = Segment(iSeg);
		iPt0 = (iSeg==iSegStart) ? max(0,iPtStart) : 0;
		iPt1 = (iSeg==iSegStop) ? min(iPtStop,pSeg->m_iSamples-1) : pSeg->m_iSamples-1;
		if( iPt1>=iPt0 ) {		// exclude when click on seg with no data - no points are selected
			iPoints += (iPt1-iPt0+1);
			//-- Calculate Sums per segment
			pSeg->CalculateSum( iChStart, iChStop, iPt0, iPt1, adAvg, iSeg==iSegStart );
			}
		}
	if( iPoints<1 ) 
		iPoints=1;
	if( piPoints != NULL ) 
		*piPoints=iPoints;
			
	//-- convert sums to averages 
	for( iCh=0; iCh<iChs; ++iCh )
		adAvg[iCh]/=iPoints;

	//-- Calculate Variance
	for( iSeg=iSegStart; iSeg<=iSegStop; ++iSeg ) { 
		CSegment * pSeg = Segment(iSeg);
		iPt0 = (iSeg==iSegStart) ? max(0,iPtStart) : 0;
		iPt1 = (iSeg==iSegStop) ? min(iPtStop,pSeg->m_iSamples-1) : pSeg->m_iSamples-1;
		if( iPt1>=iPt0 ) {		// exclude when click on seg with no data - no points are selected
			//-- calculate variances per segment
			pSeg->CalculateVariance( iChStart, iChStop, iPt0, iPt1, adAvg, adStd, iSeg==iSegStart );
			}
		}
		
	//-- convert Variances to Std dev's
	for( iCh=0; iCh<iChs; ++iCh )
		adStd[iCh] = sqrt( adStd[iCh]/iPoints );

	return true;
	}

//----------------------------------------------------------------------------------
// Get idealization event list for channel iChannel.  
// iDataSelect=0:whole file; 1:list iList; 2:current selection.	
// Returns eventlist length and fills in aiClass, aiDurations arrays 
// that must be deleted by caller.
// Each aiDuration contains number of samples, of length dSample each.
// TODO - rewrite header for stimulus settings
bool CData :: GetEventList( int iChannel, 
							int iDataSelect, int * aiCurrent, int iList, 
							bool bJoined, 
							double dDeadTime,			
							CEventList * pEventList, 
							double * pdSample			// Out : Sampling rate 
						) {

	*pdSample=m_dSampling;
	
	//-- Get Event Count
	int iEvent=0, iEvents=0;
	int iSegment, iSegments=m_aSegments.Count();
	int iRangeLow=0;
	int iRangeHigh=0;
	
	if( iDataSelect==2 ) {				// current selection 
		//-- Count events 
		for( iSegment=aiCurrent[0]; iSegment<=aiCurrent[2]; ++iSegment ) { 
			iRangeLow = iSegment==aiCurrent[0] ? aiCurrent[1] : 0;
			iRangeHigh = iSegment==aiCurrent[2] ? aiCurrent[3] : Segment(iSegment)->m_iSamples;
			iEvents += Segment(iSegment)->IDLEventList( iChannel, m_oConfig.ChannelCount(), 
							1, &iRangeLow, &iRangeHigh, m_dSampling, dDeadTime, NULL );
			}
							
		//-- allocate space for event list 
		if( ! pEventList->Allocate( iEvents, pEventList->m_iStimChannels>0 ) ) 
			return false;
			
		//-- Fill in events 
		for( iSegment=aiCurrent[0]; iSegment<=aiCurrent[2]; ++iSegment ) { //for( iSegment=0; iSegment<iSegments; ++iSegment )
			iRangeLow = iSegment==aiCurrent[0] ? aiCurrent[1] : 0;
			iRangeHigh = iSegment==aiCurrent[2] ? aiCurrent[3] : Segment(iSegment)->m_iSamples;
			CSegment * pSeg = Segment(iSegment);
			pSeg->IDLEventList( iChannel, m_oConfig.ChannelCount(), 
						1, &iRangeLow, &iRangeHigh, m_dSampling, dDeadTime,  pEventList );
			}
		}
		
	else if( iDataSelect==1 ) {			// Current List
		//-- create list of ranges from selection list with multi segment selections divided up.
		CSelList * pList = List( iList );
		if( pList==NULL ) 
			return false;
		int iSeg, iNewSeg=0, iNewSegs=0, iSel, iSelections = pList->m_aItems.Count(); 
		for( iSel=0; iSel<iSelections; ++iSel ) 
			iNewSegs += pList->Item(iSel)->m_iEndSegment - pList->Item(iSel)->m_iStartSegment;
				
		int * aiSeg = new int[iSelections];	
		int * aiStart = new int[iSelections];
		int * aiStop = new int[iSelections];
		
		for( iSel=0; iSel<iSelections; ++iSel ) { 
			CSelListItem * pSel = pList->Item( iSel );
			for( iSeg = pSel->m_iStartSegment; iSeg<=pSel->m_iEndSegment && iNewSeg<iSelections; ++iSeg ) { 
				aiSeg[iNewSeg] = iSeg;
				aiStart[iNewSeg] = iSeg==pSel->m_iStartSegment ? pSel->m_iStartPoint : 0 ;
				aiStop[iNewSeg] = iSeg==pSel->m_iEndSegment ? pSel->m_iEndPoint : 0 ;
				++iNewSeg;
				}
			}

		//-- Count events 
		iNewSeg=0;
		for( int i=0; i<iNewSegs; ++i ) { 
			if( i==iNewSegs-1 || aiSeg[i] != aiSeg[iNewSeg] ) { 
				iEvents += Segment(aiSeg[i])->IDLEventList( iChannel, m_oConfig.ChannelCount(), 
							i-iNewSeg+1, aiStart+iNewSeg, aiStop+iNewSeg, m_dSampling, dDeadTime, NULL );
				iNewSeg=i;
				}
			}
			
		//-- allocate space for event list 
		if( ! pEventList->Allocate( iEvents, pEventList->m_iStimChannels>0 ) ) 
			return false;

		//-- Fill in events 
		iNewSeg=0;
		for( int i=0; i<iNewSegs; ++i ) { 
			if( i==iNewSegs-1 || aiSeg[i] != aiSeg[iNewSeg] ) { 
				Segment(aiSeg[i])->IDLEventList( iChannel, m_oConfig.ChannelCount(), 
							i-iNewSeg+1, aiStart+iNewSeg, aiStop+iNewSeg, m_dSampling, dDeadTime, pEventList );
				iNewSeg=i;
				}
			}

		//-- clean up 
		delete [] aiSeg;
		delete [] aiStart; 
		delete [] aiStop;
		}
		
	else {								//-- Whole File 
		//-- Count events 
		for( iSegment=0; iSegment<iSegments; ++iSegment ) { 
			iRangeHigh=Segment(iSegment)->m_iSamples;
			iEvents += Segment(iSegment)->IDLEventList( iChannel, m_oConfig.ChannelCount(), 
							1, &iRangeLow, &iRangeHigh, m_dSampling, dDeadTime, NULL );
			}
		if( ! bJoined )					// count segment separation events
			iEvents+=iSegments-1;
							
		//-- allocate space for event list 
		if( ! pEventList->Allocate( iEvents, pEventList->m_iStimChannels>0 ) ) 
			return false;
			
		//-- Fill in events 
		for( iSegment=0; iSegment<iSegments; ++iSegment ) {
			//-- add a list of one range events to the overall list
			CSegment * pSeg = Segment(iSegment);
			iRangeHigh=Segment(iSegment)->m_iSamples;
			int iSegEvts = pSeg->IDLEventList( iChannel, m_oConfig.ChannelCount(), 
							1, &iRangeLow, &iRangeHigh, m_dSampling, dDeadTime,  pEventList );
			iEvent += iSegEvts;
			
			//-- add a segment separator event
			if( ! bJoined && iSegment+1<iSegments && iEvent<iEvents ) { 
				pEventList->AddEvent( 0, -1 );
				}
			}
		}

	return true;
	}
	
//======================================================================= Lists 
CSelList :: CSelList( const char * pchName ) { 
	if( pchName!=NULL )
		m_cName.Set( pchName );
	}
	
CSelListItem * CSelList :: Item( int iItem ) {
	return (CSelListItem *) m_aItems[iItem];
	}

void CSelListItem :: Text( char * pch ) { 
	sprintf( pch, "%d:%d-%d:%d %s", m_iStartSegment, m_iStartPoint, m_iEndSegment, m_iEndPoint, m_cLabel.ConstBuffer() );
	}

int CSelList :: AddListItem( int iStartSegment, int iStartPoint, int iEndSegment, int iEndPoint, const char * pchLabel, const char * pchNote ) { 
	CSelListItem * pItem;
	//-- determine sorting position of new item - first item > current, or end of list
	int iSel;
	for( iSel=0; iSel<m_aItems.Count(); ++iSel ) { 
		pItem = Item(iSel);
		if( pItem->m_iStartSegment>iStartSegment 
				|| ( pItem->m_iStartSegment==iStartSegment && pItem->m_iStartPoint>iStartPoint ) ) 
			break;
		}
	
	//-- create new item
	pItem = new CSelListItem;
	pItem->m_iStartSegment=iStartSegment;
	pItem->m_iStartPoint=iStartPoint;
	pItem->m_iEndSegment=iEndSegment;
	pItem->m_iEndPoint=iEndPoint;
	if( pchLabel != NULL ) 
		pItem->m_cLabel=pchLabel;
	if( pchNote != NULL ) 
		pItem->m_cNote=pchNote;

	//-- add new item to list at correct sorting position
	m_aItems.AddItem( pItem, iSel );
	
	return iSel;
	}

void CSelListItem :: Set( int iStartSegment, int iStartPoint, int iEndSegment, int iEndPoint ) { 
	m_iStartSegment=iStartSegment;
	m_iStartPoint=iStartPoint;
	m_iEndSegment=iEndSegment; 
	m_iEndPoint=iEndPoint;
	}

//======================================================================= Selection Sets 
int CSelectionSetRange :: Size() { 
	return m_iMax-m_iMin+1;
	}

CSelectionSet :: CSelectionSet( ){
	m_iCurrent=-1;
	m_iMaxSize=0;
	m_iCurrentSize=0;
	m_adData=NULL;
	m_pData=NULL;
	m_aiClass=NULL;
	m_aiState=NULL;
	m_iAllocatedIDL=0;
	}

CSelectionSet :: ~CSelectionSet(){
	if( m_adData!=NULL ) 
		delete [] m_adData;
	if( m_aiClass!=NULL ) 
		delete [] m_aiClass;
	if( m_aiState!=NULL ) 
		delete [] m_aiState;
	}

//-- Scan Ranges for maximum size to use for buffer allocation 
void CSelectionSet :: AllocateBuffer( int iChannels ) { 
	m_iMaxSize=0;
	
	CSelectionSetRange * pRange;
	for( int i=0; i<m_aRanges.Count(); ++i ) {
		pRange = (CSelectionSetRange *) m_aRanges[i];
		m_iMaxSize=max(m_iMaxSize,pRange->Size());
		}
		
	if( m_adData != NULL ) 
		delete [] m_adData;
	m_adData = new double[m_iMaxSize*iChannels];
	}

void CSelectionSet :: SetupHelper( int iSeg0, int iOffset0, int iSeg1, int iOffset1 ) { 
	CSelectionSetRange * pRange;
	while( iSeg0 <= iSeg1 ) { 
		CSegment * pSegment = m_pData->Segment(iSeg0);
		m_aRanges.AddItem( pRange = new CSelectionSetRange );
		pRange->m_dtStart0=pSegment->m_dtStartTime;
		pRange->m_dtStart=pSegment->m_dtStartTime+m_pData->m_dSampling*iOffset0;
		pRange->m_iSeg=iSeg0;
		pRange->m_iMin=iOffset0;
		pRange->m_iMax = (iSeg0==iSeg1?iOffset1:(pSegment->m_iSamples-1) );
		++iSeg0;
		iOffset0=0;
		}
	}
	
//-- Call one of the three setup's 
bool CSelectionSet :: Setup( HWND hwndData, int iMode, CData * pData, int iChannel ) { 
	if( iMode==0 )			//- whole file 
		return Setup( pData, iChannel );

	if( iMode==1 ) {			//- current list
		int iList;
		((CDialog *) GetWindowLong( GetParent(hwndData), 0 ))->GetData( 4, &iList );
		if( iList<0 ) {
			MessageBox( GetParent(hwndData), "Selected data file does not have a current list", "Load Data", MB_OK );
			return false;
			}
		return Setup( pData, iChannel, iList );
		}

	if( iMode==2 ) { 
		int ai[4];
		((CDialog *) GetWindowLong( GetParent(hwndData), 0 ))->GetData( 3, ai );
		return Setup( pData, iChannel, ai[0], ai[1], ai[2], ai[3] );
		}
	return false;
	}

//-- whole file selection set
bool CSelectionSet :: Setup( CData * pData, int iChannel ){
	m_pData=pData;
	m_iChannel=iChannel;
	int iSegs = pData->m_aSegments.Count();
	SetupHelper( 0, 0, iSegs-1, pData->Segment(iSegs-1)->m_iSamples-1 );
	AllocateBuffer( iChannel==-1?pData->m_oConfig.ChannelCount():1 );		
	return true;
	}

//-- constructor for list selection set 
bool CSelectionSet :: Setup( CData * pData, int iChannel, int iList ){
	m_pData=pData;
	m_iChannel=iChannel;
	m_iMaxSize=0;
	if( iList<0 || iList>=m_pData->m_aLists.Count() ) 
		return false;
	
	CSelList * pList = m_pData->List(iList);
	CSelListItem * pItem;
	for( int iItem=0; iItem<pList->m_aItems.Count(); ++iItem ) { 
		pItem = pList->Item(iItem);
		SetupHelper( pItem->m_iStartSegment, pItem->m_iStartPoint, pItem->m_iEndSegment, pItem->m_iEndPoint );
		}

	AllocateBuffer( iChannel==-1?pData->m_oConfig.ChannelCount():1 );		
	return true;
	}

bool CSelectionSet :: Setup( CData * pData, int iChannel, const char * pchList ){
	m_pData=pData;
	m_iChannel=iChannel;
	m_iMaxSize=0;
	for( int iList=0; iList<pData->m_aLists.Count(); ++iList ) { 
		if( strcmp(((CSelList *) m_pData->List(iList))->m_cName.Buffer(),pchList)==0 ) {
			Setup( pData, iChannel, iList );
			return true;
			}
		}
	return false;
	}

//-- constructor for current highlight selection list
bool CSelectionSet :: Setup( CData * pData, int iChannel, int iSeg0, int iOffset0, int iSeg1, int iOffset1 ){
	m_pData=pData;
	m_iChannel=iChannel;
	SetupHelper( iSeg0, iOffset0, iSeg1, iOffset1 );
	AllocateBuffer( iChannel==-1?pData->m_oConfig.ChannelCount():1 );		
	return true;
	}

int CSelectionSet :: RangeIndex() { 
	return m_iCurrent;
	}

int CSelectionSet :: CurrentSize() { 
	return m_iCurrentSize;
	}

bool CSelectionSet :: RangeMinMax( int iRange, int iChannel, bool bInitialize, double * pdMin, double * pdMax ) { 
	//-- check if iRange is valid
	if( iRange>=m_aRanges.Count() || iRange<0 ) {
		m_iCurrentSize=0;
		return false;
		}

	//-- new current data range
	CSelectionSetRange * pRange = (CSelectionSetRange *)(m_aRanges[iRange]);
	int iSize = pRange->Size();

	//-- Make sure data has been loaded to CData 
	CSegment * pSeg = m_pData->Segment( pRange->m_iSeg );
	if( pSeg->m_pdData==NULL ) // check here, not in Load..() since m_pdDataFile may be NULL for new data.
		m_pData->m_pDataFile->LoadSegmentData( pSeg );
	
	//-- copy from CData to local buffer 
	double d;
	int iChannels = m_pData->m_oConfig.ChannelCount();
	if( bInitialize )
		*pdMax=*pdMin=pSeg->m_pdData[pRange->m_iMin*iChannels+iChannel];
	for( int i=pRange->m_iMin; i<=pRange->m_iMax; ++i ) {
		d = pSeg->m_pdData[i*iChannels+m_iChannel];
		if( *pdMin>d ) 
			*pdMin=d;
		if( *pdMax<d ) 
			*pdMax=d;
		}
		
	return true;
	}

bool CSelectionSet :: CalculateMinMax( int iChannel, double * pdMin, double * pdMax ) { 
	bool bFirst=true;
	int iRanges=m_aRanges.Count();
	for( int iRange=0; iRange<iRanges; ++iRange ) { 
		if( ((CSelectionSetRange *)(m_aRanges[iRange]))->Size()>0 ) {
			RangeMinMax( iRange, iChannel, bFirst, pdMin, pdMax );
			bFirst=false;
			}
		}
	return true;
	}
	
bool CSelectionSet :: GetRange( int iRange ){
	//-- check if iRange is valid
	if( iRange>=m_aRanges.Count() || iRange<0 ) {
		m_iCurrentSize=0;
		return false;
		}

	//-- new current data range
	m_pRange = (CSelectionSetRange *)(m_aRanges[m_iCurrent=iRange]);
	m_iCurrentSize=m_pRange->Size();

	//-- Make sure data has been loaded to CData 
	CSegment * pSeg = m_pData->Segment( m_pRange->m_iSeg );
	if( pSeg->m_pdData==NULL ) // check here, not in Load..() since m_pdDataFile may be NULL for new data.
		m_pData->m_pDataFile->LoadSegmentData( pSeg );
	
	//-- copy from CData to local buffer 
	int iChannels = m_pData->m_oConfig.ChannelCount();
	if( m_iChannel==-1 )
		memcpy( m_adData, pSeg->m_pdData+iChannels*m_pRange->m_iMin, iChannels*m_pRange->Size()*sizeof(double) );
	else
		for( int i=m_pRange->m_iMin; i<=m_pRange->m_iMax; ++i ) 
			m_adData[i-m_pRange->m_iMin]=pSeg->m_pdData[i*iChannels+m_iChannel];
		
	return true;
	}

double CSelectionSet :: Datum( int iSample ) { 
	return m_adData[iSample];
	}
	
double CSelectionSet :: Datum( int iChannel, int iSample ) { 
	if( m_iChannel==-1 ) 
		return m_adData[iSample*(m_pData->m_oConfig.ChannelCount())+iChannel];
	return m_adData[iSample];
	}

bool CSelectionSet :: GetFirst(){
	return GetRange(0);
	}

bool CSelectionSet :: GetNext(){
	return GetRange( ++m_iCurrent );
	}

//------------------------ ---------- ------ --------- ----------
bool CSelectionSet :: GetIdealized() { 
	//-- check for current non-empty selection set range
	int iLen = m_pRange->Size();
	if( m_pRange == NULL || iLen<=0 || m_iMaxSize<=0 ) 
		return false;

	//-- Allocate space for IDL, clear previous if necc.
	if( m_iAllocatedIDL < iLen ) { 
		m_iAllocatedIDL=max(m_iMaxSize,iLen);
		if( m_aiClass != NULL ) 
			delete[]m_aiClass; 
		m_aiClass = new short[ m_iAllocatedIDL ];

		if( m_aiState != NULL ) 
			delete[]m_aiState; 
		m_aiState = new short[ m_iAllocatedIDL ];
		
		if( m_aiClass==NULL || m_aiState==NULL ) {
			if( m_aiClass!=NULL ) 
				delete [] m_aiClass;
			m_aiClass=m_aiState=NULL;
			m_iAllocatedIDL=0;
			}
		}

	//-- access data segment	
	CSegment * pSeg = m_pData->Segment(m_pRange->m_iSeg);

	//-- Scan IDL's
	bool bFound=true;
	for( int iIdlList=0; iIdlList<pSeg->m_aIdl.Count(); ++iIdlList ) {
		CIdlList * pIdl = pSeg->IdlList(iIdlList);
		if( pIdl->m_iOffset+pIdl->m_iLength-1 >= m_pRange->m_iMin && pIdl->m_iOffset <= m_pRange->m_iMax ) { 
			//-- some or all of this IDL applies to desired range.  Scan dwells.
			int iOffset=pIdl->m_iOffset;			// accumulate offset to start of dwell
			int iNext;								// calculate start of next segment (if any)
			for( int iDwell=0; iDwell<pIdl->m_aIdlDwells.Count(); ++iDwell ) { 
				CIdlDwell * pDwell = pIdl->IdlDwell( iDwell );
				iNext=iOffset+pDwell->m_iDuration;
				for( int iPt=max(iOffset,m_pRange->m_iMin); iPt<=min(iNext-1,m_pRange->m_iMax); ++iPt ) { 
					m_aiClass[iPt-m_pRange->m_iMin]=pDwell->m_iClass;
					m_aiState[iPt-m_pRange->m_iMin]=pDwell->m_iState;
					}
				iOffset=iNext;
				}
			}
		}
	return true;
	}

CSelectionSetRange * CSelectionSet :: Range( int iRange ) { 
	return (CSelectionSetRange *) (m_aRanges[iRange]);
	} 

int CSelectionSet :: MaxSize() { 
	return m_iMaxSize;
	}

int CSelectionSet :: TotalSize() { 
	int iSize=0;
	for( int i=0; i<m_aRanges.Count(); ++i ) 
		iSize += Range(i)->Size();
	return iSize;	
	}
